/** @file CalValsTool.cxx
@brief Calculates the Cal analysis variables
@author Bill Atwood, Leon Rochester

$Header: /nfs/slac/g/glast/ground/cvs/AnalysisNtuple/src/CalValsTool.cxx,v 1.82 2007/03/14 22:20:05 lsrea Exp $
*/
//#define PRE_CALMOD 1

// Include files


#include "ValBase.h"

#include "GaudiKernel/MsgStream.h"
#include "GaudiKernel/IDataProviderSvc.h"
#include "GaudiKernel/SmartDataPtr.h"
#include "GaudiKernel/ToolFactory.h"
#include "GaudiKernel/IToolSvc.h"

#include "Event/TopLevel/EventModel.h"
#include "Event/TopLevel/Event.h"

#include "Event/Recon/TkrRecon/TkrCluster.h"
#include "Event/Recon/TkrRecon/TkrTrack.h"
#include "Event/Recon/TkrRecon/TkrVertex.h"
#include "Event/Recon/CalRecon/CalCluster.h"
#include "Event/Recon/CalRecon/CalEventEnergy.h"
#include "Event/Recon/CalRecon/CalParams.h"
#include "Event/Recon/CalRecon/CalXtalRecData.h"

#include "GaudiKernel/IToolSvc.h"
//#include "GlastSvc/Reco/IPropagatorTool.h"
//#include "GlastSvc/Reco/IPropagator.h" 

#include "GlastSvc/GlastDetSvc/IGlastDetSvc.h"
#include "GlastSvc/Reco/IPropagatorSvc.h"
#include "TkrUtil/ITkrGeometrySvc.h"

#include "idents/TowerId.h" 
#include "idents/VolumeIdentifier.h"

#include "CLHEP/Geometry/Transform3D.h"
#include "CLHEP/Vector/Rotation.h"

#include "TMath.h"

#include "Doca.h"

/*! @class CalValsTool
@brief calculates Cal values

@authors Bill Atwood, Leon Rochester
*/

namespace {
    double truncFraction = 0.9;
}

class CalValsTool :   public ValBase
{
public:

    CalValsTool( const std::string& type, 
        const std::string& name, 
        const IInterface* parent);

    virtual ~CalValsTool() { }

    StatusCode initialize();

    StatusCode calculate();

private:

    // some pointers to services  
    /// GlastDetSvc used for access to detector info
    IGlastDetSvc*    m_detSvc; 
    /// TkrGeometrySvc used for access to tracker geometry info
    ITkrGeometrySvc* m_tkrGeom;

    IPropagator*     m_G4PropTool;    

    /// some Geometry
    double m_towerPitch;
    int    m_xNum;
    int    m_yNum;
    int    m_nLayers;
    int    m_nCsI;

    /// gets the CAL info from detModel
    StatusCode getCalInfo();
    double activeDist(Point pos, int &view) const;


};
namespace {
    // this is the test distance for the CAL_EdgeEnergy variable
    double _deltaEdge = 50.0;
}

// Static factory for instantiation of algtool objects
static ToolFactory<CalValsTool> s_factory;
const IToolFactory& CalValsToolFactory = s_factory;

// Standard Constructor
CalValsTool::CalValsTool(const std::string& type, 
                         const std::string& name, 
                         const IInterface* parent)
                         : ValBase( type, name, parent )
{    
    // Declare additional interface
    declareInterface<IValsTool>(this); 
}

StatusCode CalValsTool::initialize()
{
    StatusCode sc = StatusCode::SUCCESS;


    // load up the map

    /** @page anatup_vars 
    @section calvalstool CalValsTool Variables
    <table>
    <tr><th> Variable <th> Type <th> Description
    <tr><td> CalEnergyRaw 
    <td>F<td>   Sum of the raw energies in all the crystals.  
    Includes estimate of missed energy due to zero-supression.  
    This replaces the variable CalEnergySum.  NEW! 
    <tr><td> CalEnergyCorr 
    <td>F<td>   Cal Energy corrected layer-by-layer for edges and leakage.  
    This replaces the variable CalEneSumCorr. NEW! 
    <tr><td> CalLeakCorr 
    <td>F<td>   Leakage correction: this is the contained fraction of the total energy 
    after edge corrections.  
    <tr><td> CalEdgeCorr 
    <td>F<td>   Effective layer-by-layer edge correction mainly due to the gaps 
    between Cal modules; multiplicative 
    <tr><td> CalTotalCorr 
    <td>F<td>   Global total correction. Includes effect due to dead material; 
    multiplicative 
    <tr><td> CalCsIRLn 
    <td>F<td>   Total radiation lengths in crystals, integrated along the 
    event axis (line connecting the first hit in the tracker to the CAL energy centroid) 
    <tr><td> CalTotRLn 
    <td>F<td>   Total radiation lengths in the CAL, integrated along the event axis. 
    <tr><td> CalCntRLn 
    <td>F<td>   Radiation lengths integrated along the event axis, up to energy centroid 
    <tr><td> CalLATRLn 
    <td>F<td>   Total radiation lengths integrated along the event axis 
    (including the tracker). 
    <tr><td> CalDeadTotRat 
    <td>F<td>   Ratio of radiation lengths in dead material to CalTotRLn 
    <tr><td> CalDeadCntRat 
    <td>F<td>   Ratio of radiation lengths in dead material up to energy centroid, 
    to CalCntRat 
    <tr><td> CalTPred 
    <td>F<td>   Model-predicted energy centroid in radiation lengths 
    <tr><td> CalDeltaT 
    <td>F<td>   Difference between measured and predicted energy centroids 
    <tr><td> CalTwrEdge 
    <td>F<td>   Distance of the entry point of the best track from the tower boundary, 
    measured at the top of the CAL. 
    <tr><td> CalLATEdge 
    <td>F<td>   Closest distance of track 1, projected to the top of the CAL, 
    to the edge of the CAL layer, taking non-square shape into account. 
    This is essentially the old merit skirt variable. 
    <tr><td> CalEdgeEnergy 
    <td>F<td>   The sum of the raw energies in each crystal for which the energy centroid 
    is within _deltaEdge (currently 50 mm) of the outside edge of one of 
    the outside CAL modules.
    This is an attempt at a "anti-coincidence counter" for the CAL.
    <tr><td> CalTwrEdgeCntr 
    <td>F<td>   Distance of the energy centroid from the nearest tower boundary.  
    <tr><td> CalGapFraction 
    <td>F<td>   Approximate fraction of the shower volumn which falls in inter-tower gaps. 
    <tr><td> CalTrackSep 
    <td>F<td>   Distance between impact points of two best tracks at CAL front face; 
    zero if only one track 
    <tr><td> CalTrackDoca 
    <td>F<td>   Distance between the projected vertex (or track if only one track) 
    and the energy centroid, evaluated at the z of the centroid. 
    <tr><td> CalTrackAngle 
    <td>F<td>   Angle between "gamma" direction in the tracker and direction of the CAL "track" 
    <tr><td> CalELayerN, N=0,7 
    <td>F<td>   Energy deposited in layer N of the CAL 
    <tr><td> CalLyr0Ratio 
    <td>F<td>   Ratio of CalELayer0 to CalEnergyRaw 
    <tr><td> CalLyr7Ratio 
    <td>F<td>   Ratio of CalELayer7 to CalEnergyRaw 
    <tr><td> CalBkHalfRatio 
    <td>F<td>   Ratio of total energy in back half of CAL (layers 4-7) to 
    CalEnergyRaw 
    <tr><td> CalXtalsTrunc 
    <td>F<td>   Number of CAL Xtals with > %1 of CalEnergyRaw (see CalXtalRatio) 
    <tr><td> CalXtalRatio 
    <td>F<td>   Ratio of number of Xtals with energy > 1% of CalEnergyRaw to 
    total number of struck Xtals in the event. 
    <tr><td> CalXtalMaxEne 
    <td>F<td>   Maximum energy found in a single Xtal
    <tr><td> CalLongRms 
    <td>F<td>   rms of the average of the 1st and 3rd shower moments. 
    Indicates the length of the measured shower along the shower axis. 
    <tr><td> CalLRmsAsym 
    <td>F<td>   The asymetry of the 1st and 3rd shower moments.  
    This should be close to zero. Because of ordering of moments it is slightly ... (??)
    <tr><td> CalTransRms 
    <td>F<td>   rms of transverse position measurements.
    <tr><td> CalMIPDiff 
    <td>F<td>   Difference between measured energy and that expected 
    from a minimum-ionizing particle 
    <tr><td> CalMIPRatio 
    <td>F<td>   Ratio of measured energy to that expected from a 
    minimum-ionizing particle 
    <tr><td> Cal[X/Y/Z]Ecntr 
    <td>F<td>   Energy centroid in [x/y/z]
    <tr><td> Cal[X/Y/Z]Dir 
    <td>F<td>   [x/y/z] direction cosine of CAL "track" 
    <tr><td> Cal[X/Y]0 
    <td>F<td>   [x/y] position of CAL "track" measured at the energy centroid 
    <tr><td> CalTrkXtalRms
    <td>F<td>   For this and the next three variables, a measure of the rms spread
    of the crystals in the calorimeter around the projection of the first
    track. They differ by the weighting and crystal count. This one is 
    the rms of the DOCAs of all crystals with repect to the projected track
    <tr><td> CalTrkXtalRmsE
    <td>F<td>   Same as the previous, but weighted by the deposited energy in 
    each crystal
    <tr><td> CalTrkXtalRmsTrunc
    <td>F<td>   Same as the first, but excludes the crystals with the largest DOCAs.
    Default is to remove 10% of the crystals, with the number of crystals
    included rounded to the nearest integer
    <tr><td> CalTrkXtalRmsETrunc
    <td>F<td>   Same as the previous, but weighted by the deposited energy in
    each crystal
    <tr><td> CalRmsLayerE
    <td>F<td>   Rms of deposited energy (normalized to rad. lengths traversed, 
    and to average energy deposited) for all
    the layers with > 5% of deposited raw energy and at least 0.5 rad lengths
    of predicted CsI traversed. 
    <tr><td> CalRmsLayerEBack
    <td>F<td>   Same as above, with layer 0 left out of the calculation
    <tr><td> CalNLayersRmsBacj
    <td>I<td>    Number of layers used in the calculation of CalRmsLayerEBack above
    <tr><td> CalEAveBack
    <td>F<td>   Average of normalized eDep excluding layer 0, cuts as for CalRmsLayerE above
    <tr><td> CalLayer0Ratio
    <td>F<td>   Ratio of layer0 normalized eDep to the average of the remaining layers, 
    cuts as for CalRmsLayerE above
    <tr><td> Cal[X/Y]PosRmsLL
    <td>F<td>   Energy Weighted Rms of the hit crystals in the last layer
    (layer 7). X is measured across the width of the crystals,
    Y across the length.
    </table>

    */

    return sc;
}

StatusCode CalValsTool::calculate()
{
    StatusCode sc = StatusCode::SUCCESS;


    return sc;
}

